/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  AFILE *AFopenWrite (const char Fname[], int Fformat, long int Nchan,
                      double Sfreq, FILE *fpinfo)

Purpose:
  Open an audio file for writing

Description:
  This routine opens an audio file for writing.  This routine sets up the audio
  file parameters to write data of the given format to the audio file.  After
  writing data to the file, the routine AFclose should be called to update the
  file header information and close the file.

  By default, information consisting of the date, the user and the program
  creating the file is written to the audio file header.  The routine
  AFsetHInfo can be called before calling this routine to specify additional
  information to be written to the header.  Any such additional header
  information is reset to null after being written to the header.
 
  This routine can write AU audio files, AIFF-C files, RIFF WAVE files, and
  headerless audio files.

  AU audio file:
    8-bit mu-law, 8-bit A-law, 8-bit integer, 16-bit integer, 24-bit integer,
    32-bit integer, and 32-bit IEEE floating-point, and 64-bit IEEE
    floating-point data formats are supported.
  RIFF WAVE file:
    8-bit mu-law, 8-bit A-law, offset-binary 8-bit integer, and 16-bit integer,
    24-bit integer, 32-bit integer, 32-bit IEEE floating-point, and 64-bit IEEE
    floating-pont data formats are supported.
  AIFF-C file:
    8-bit mu-law, 8-bit A-law, 8-bit integer, 16-bit integer, 24-bit integer,
    and 32-bit integer, 32-bit IEEE floating-point, and 64-bit IEEE
    floating-point data formats are supported.
  Headerless file:
    8-bit mu-law, 8-bit A-law, offset-binary 8-bit integer, 8-bit integer,
    16-bit integer, 24-bit integer, 32-bit integer, and 32-bit floating-point,
    64-bit floating-point, and text data formats are supported.  A text format
    file has the data in character form, one value to a line.

  For the fixed point file data representations, input values in the following
  ranges will be converted without clipping.
     data format       allowed values
    8-bit mu-law   - [ -32636, +32636 ]
    8-bit A-law    - [ -32768, +32768 ]
    8-bit integer  - [ -16384, -16383 ]
    16-bit integer - [ -32768, +32767 ]
    24-bit integer - [ -32768, +32768 )
    32-bit integer - [ -32768, +32768 )
  For AU and WAVE files containing floating-point data, values are scaled
  by 1/32768 before being written to the file.  For data values between
  [-32768, +32768], the file data will lie in the range [-1, +1].  For
  headerless files of any data format, the values are written scaled by unity.

Parameters:
  <-  AFILE *AFopenWrite
      Audio file pointer for the audio file
   -> const char Fname[]
      Character string specifying the file name.  The file name "-" means
      standard output.
   -> int Fformat
      Audio file format code, evaluated as the sum of a data format code and a
      file type,
        Fformat = Dformat + 256 * Ftype
      For Ftype equal to zero, the file will have the standard AU audio file
      header.  The Ftype flag allows for other file types as described below.
      Note that not all data formats are allowed for all file types.
      Dformat: data format
        FD_MULAW8  = 1,  mu-law 8-bit data
        FD_ALAW8   = 2,  A-law 8-bit data
        FD_UINT8   = 3,  offset-binary 8-bit integer data
        FD_INT8    = 4,  two's-complement 8-bit integer data
        FD_INT16   = 5,  two's-complement 16-bit integer data
        FD_INT24   = 6,  two's-complement 24-bit integer data
        FD_INT32   = 7,  two's-complement 32-bit integer data
        FD_FLOAT32 = 8,  32-bit floating-point data
        FD_FLOAT64 = 9,  64-bit floating-point data
        FD_TEXT    = 10,  text data
      Ftype: file type
        FTW_DEFAULT   = 0, default audio file type
        FTW_AU        = 1, AU audio file
        FTW_WAVE      = 2, RIFF WAVE file
        FTW_AIFF_C    = 3, AIFF-C audio file
        FTW_NH_EB     = 4, Headerless file (big-endian byte order)
        FTW_NH_EL     = 20, Headerless file (little-endian byte order)
        FTW_NH_NATIVE = 36, Headerless file (native byte order)
        FTW_NH_SWAP   = 52, Headerless file (byte-swapped byte order)
   -> long int Nchan
      Number of channels
   -> double Sfreq
      Sampling frequency
   -> FILE *fpinfo
      File pointer for printing audio file information.  If fpinfo is not NULL,
      information about the audio file is printed on the stream selected by
      fpinfo.

Author / revision:
  P. Kabal  Copyright (C) 1999

  $Revision: 1.68 $  $Date: 1999/06/14 14:16:32 $

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: AFopenWrite.c 1.68 1999/06/14 AFsp-v5r2 $";

#include <stdio.h>
#include <stdlib.h>		/* EXIT_FAILURE */
#include <string.h>

#include <libtsp.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#define AF_DATA_TYPE_NAMES
#include <libtsp/AFpar.h>

#define ERR_MSG	0
#define SYS_MSG	1

static int
AF_checkDT (int Ftype, int Dformat);
static int
AF_checkPar (int Ftype, int Dformat, long int Nframe, long int Nchan);
static void
AF_error (const char Fname[], int sysFlag);
static FILE*
AF_open_write (const char Fname[], int Dformat);
static void
AF_resetPar (void);

/* If AFopts_ErrorHalt is clear, execution continues after an error */


AFILE *
AFopenWrite (const char Fname[], int Fformat, long int Nchan, double Sfreq,
	     FILE *fpinfo)

{
  AFILE *AFp;
  FILE *fp;
  int Dformat, Ftype, Swapb;
  long int Nframe;

/* Separate the file type and data type */
  Dformat = FTW_dformat (Fformat);
  Ftype = FTW_ftype (Fformat);

/* Check the input parameters */
  Nframe = AFopt_Nframe;
  if (AF_checkPar (Ftype, Dformat, Nframe, Nchan)) {
    AF_error (Fname, ERR_MSG);
    return NULL;
  }

/* Open the file for writing */
  fp = AF_open_write (Fname, Dformat);
  if (fp == NULL)
    return NULL;

/* Set up the audio file parameters and write the header information */
  switch (Ftype) {
  case FTW_AU:
    AFp = AFwrAUhead (fp, Dformat, Nframe, Nchan, Sfreq);
    break;
  case FTW_WAVE:
    AFp = AFwrWVhead (fp, Dformat, Nframe, Nchan, Sfreq);
    break;
  case FTW_AIFF_C:
    AFp = AFwrAIhead (fp, Dformat, Nframe, Nchan, Sfreq);
    break;
  case FTW_NH_EB:
  case FTW_NH_EL:
  case FTW_NH_NATIVE:
  case FTW_NH_SWAP:
    Swapb = FTW_subtype (Ftype);
    AFp = AFsetNHwrite (fp, Dformat, Swapb, Nchan, Sfreq);
    break;
  default:
    AFp = NULL;
    break;
  }

/* Error messages */
  if (AFp == NULL) {
    fclose (fp);
    AF_error (Fname, ERR_MSG);
    return NULL;
  }

/* Reset open parameters */
  AF_resetPar ();

/* Print the header information */
  AFprAFpar (AFp, Fname, fpinfo);

return AFp;
}

/* Check the user supplied parameters */


static int
AF_checkPar (int Ftype, int Dformat, long int Nframe, long int Nchan)

{
  int ErrCode;

  /* Check the file type / data format combination */
  ErrCode = AF_checkDT (Ftype, Dformat);

  if (Nchan <= 0L) {
    UTwarn ("AFopenWrite - %s: %ld", AFM_BadNChan, Nchan);
    ErrCode = 1;
  }
  if (Nframe != AF_NFRAME_UNDEF && Nframe < 0L) {
    UTwarn ("AFopenWrite - %s: %ld", AFM_BadNFrame, Nframe);
    ErrCode = 1;
  }

  return ErrCode;
}

/* Check for file type / data type compatibility */


static int
AF_checkDT (int Ftype, int Dformat)

{
  int ErrCode;

/* undef mulaw  Alaw uint8  int8 int16 int24 int32 flt32 flt64 text */
  static const int DT_AU[NFD] = {
     0,    1,    1,    0,    1,    1,    1,    1,    1,    1,    0};
  static const int DT_WAVE[NFD] = {
     0,    1,    1,    1,    0,    1,    1,    1,    1,    1,    0};
  static const int DT_AIFF_C[NFD] = {
     0,    1,    1,    0,    1,    1,    1,    1,    1,    1,    0};

  ErrCode = 0;
  if (Dformat < 1 || Dformat >= NFD) {
    UTwarn ("AFopenWrite - %s: \"%d\"", AFM_BadDataC, Dformat);
    ErrCode = 1;
    return ErrCode;
  }

  switch (Ftype) {
  case FTW_AU:
    if (DT_AU[Dformat] == 0) {
      UTwarn ("AFopenWrite - %s: \"%s\"", AFM_AU_UnsData, AF_DTN[Dformat]);
      ErrCode = 1;
    }
    break;
  case FTW_WAVE:
    if (DT_WAVE[Dformat] == 0) {
      UTwarn ("AFopenWrite - %s: \"%s\"", AFM_WV_UnsData, AF_DTN[Dformat]);
      ErrCode = 1;
    }
    break;
  case FTW_AIFF_C:
    if (DT_AIFF_C[Dformat] == 0) {
      UTwarn ("AFopenWrite - %s: \"%s\"", AFM_AIFF_UnsData, AF_DTN[Dformat]);
      ErrCode = 1;
    }
    break;
  case FTW_NH_EB:
  case FTW_NH_EL:
  case FTW_NH_NATIVE:
  case FTW_NH_SWAP:
    break;		/* All data types supported */
  case FTW_UNDEF:
    UTwarn ("AFopenWrite - %s", AFM_NoFType);
    ErrCode = 1;
    break;
  default:
    UTwarn ("AFopenWrite - %s: %d", AFM_BadFTypeC, Ftype);
    ErrCode = 1;
    break;
  }

  return ErrCode;
}

/* Open a file for writing */

#include <libtsp/sysOS.h>
#if (SY_OS == SY_OS_WINDOWS || SY_OS == SY_OS_CYGWIN)
#  include <io.h>	/* setmode */
#  include <fcntl.h>	/* IO_BINARY */
#endif

static FILE *
AF_open_write (const char Fname[], int Dformat)

{
  FILE *fp;

  if (strcmp (Fname, "-") == 0) {

    /* Output to standard output */
    fp = stdout;
#if (SY_OS == SY_OS_WINDOWS || SY_OS == SY_OS_CYGWIN)
    if (Dformat == FD_TEXT) {
      if (setmode (fileno(fp), O_TEXT) == -1)
	UTwarn ("AFopenRead - %s", AFM_NoWTextMode);
    }
    else {
      if (setmode (fileno(fp), O_BINARY) == -1)
	UTwarn ("AFopenRead - %s", AFM_NoWBinMode);
    }
#endif

  }
  else {

    /* Output to a file */
    if (Dformat == FD_TEXT)
      fp = fopen (Fname, "w");
    else
      fp = fopen (Fname, "wb");
  }

  if (fp == NULL) {
    AF_error (Fname, SYS_MSG);
    return NULL;
  }

  return fp;
}

/* Print an error message with the file name */


static void
AF_error (const char Fname[], int sysFlag)

{
  const char *fn;

  if (strcmp (Fname, "-") == 0)
    fn = "<stdout>";
  else
    fn = Fname;

  if (sysFlag)
    UTsysMsg ("AFopenWrite: %s \"%s\"", AFM_OpenWErr, fn);
  else
    UTwarn ("AFopenWrite: %s \"%s\"", AFM_OpenWErr, fn);
  if (AFopt_ErrorHalt)
    exit (EXIT_FAILURE);

  /* Reset open parameters */
  AF_resetPar ();

  return;
}

/* Reset per file parmeters */


static void
AF_resetPar (void)

{
  AFopt_Nframe = AF_NFRAME_UNDEF;
  AFsetHinfo (NULL);

  return;
}
