/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  AFILE *AFwrAIhead (FILE *fp, int Format, long int Nframe, long int Nchan,
                     double Sfreq)

Purpose:
  Write header information to an AIFF-C file

Description:
  This routine opens and writes header information to an AIFF-C format audio
  file.

  AIFF-C audio file:
   Offset Length Type    Contents
      0     4    char   File identifier ("FORM")
      4     4    int    Chunk length
      8     4    char   File identifier ("AIFF" or "AIFC")
     12     4    char   "COMM" chunk identifier
     16     4    int    Chunk length
     24     2    int      Number of interleaved channels
     26     4    int      Number of sample frames
     30     2    int      Bits per sample
     32    10    float    Sample frames per second
     42     4    int      Compression type
     46   ...    char     Compression name
    ...    ...   ...    ...
      S     4    char   SSND chunk identifier ("SSND")
     +4     4    int    Chunk length
     +8     4    int      Data offset (0)
    +12     4    int      Block size (0)
    +16   ...    ...      Audio data

Parameters:
  <-  AFILE AFwrAIhead
      Audio file pointer for the audio file.  In case of error, the audio file
      pointer is set to NULL.
   -> FILE *fp
      File pointer for the audio file
   -> int Format
      Audio file data format code: FD_MULAW8, FD_ALAW8, FD_INT8, FD_INT16, etc.
   -> long int Nframe
      Number of sample frames.  This value is used to set the file header
      information for non-random access files for which the header cannot be
      updated when the file is closed.  If this value is AF_NFRAME_UNDEF, the
      file must be random access.
   -> long int Nchan
      Number of channels
   -> double Sfreq
      Sampling frequency

Author / revision:
  P. Kabal  Copyright (C) 2001

  $Revision: 1.32 $  $Date: 2001/04/10 01:24:59 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFwrAIhead.c 1.32 2001/04/10 AFsp-v5r2 $";

#include <assert.h>
#include <setjmp.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/nucleus.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#define AF_DATA_LENGTHS
#include <libtsp/AFpar.h>
#include <libtsp/AIpar.h>

#define ICEILV(n,m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define RNDUPV(n,m)	((m) * ICEILV (n, m))		/* Round up */

#define MCOPY(src,dest)		memcpy ((void *) (dest), (void *) (src), \
					sizeof (dest))
#define WRPAD(fp,size,align) \
     AFwriteHead (fp, NULL, 1, (int) (RNDUPV(size, align) - (size)), \
		  DS_NATIVE);

#define ALIGN		2	/* Chunks padded out to a multiple of ALIGN */

/* setjmp / longjmp environment */
extern jmp_buf AFW_JMPENV;

static void
AF_setCOMM (struct AI_CkCOMM *CkCOMM, int Format, long int Nframe,
	    long int Nchan, double Sfreq);
static double
AI_ScaleF (int Format);
static int
AF_wPstring (FILE *fp, char string[]);


AFILE *
AFwrAIhead (FILE *fp, int Format, long int Nframe, long int Nchan,
	    double Sfreq)

{
  AFILE *AFp;
  long int size, Ldata;
  struct AI_CkFORM CkFORM;
  struct AF_write AFw;

/* Set the long jump environment; on error return a NULL */
  if (setjmp (AFW_JMPENV))
    return NULL;	/* Return from a header write error */

/* Set up the encoding parameters */
  AFw.DFormat.ScaleF = AI_ScaleF (Format);
  if (AFw.DFormat.ScaleF <= 0.0)
    return NULL;
  AFw.DFormat.Format = Format;
  AFw.DFormat.Swapb = DS_EB;
  AFw.Sfreq = Sfreq;
  AFw.Nchan = Nchan;

  if (Nframe != AF_NFRAME_UNDEF)
    Ldata = Nframe * AF_DL[Format];
  else if (FLseekable (fp))
    Ldata = 0L;
  else {
    UTwarn ("AFwrAIhead - %s", AFM_AIFF_WRAccess);
    return NULL;
  }

  /* FORM chunk */
  MCOPY (FM_IFF, CkFORM.ckID);
  /* defer filling in the chunk size */
  MCOPY (FM_AIFF_C, CkFORM.AIFFID);

  /* COMM chunk */
  AF_setCOMM (&CkFORM.CkCOMM, Format, Nframe, Nchan, Sfreq);

  /* FVER chunk */
  MCOPY (CKID_FVER, CkFORM.CkFVER.ckID);
  CkFORM.CkFVER.ckSize = 4;
  CkFORM.CkFVER.timestamp = AIFCVersion1;

  /* Header information records */
  AFw.Hinfo = *(AFgenHinfo (0.0));

  /* ANNO chunk */
  MCOPY (CKID_ANNOTATION, CkFORM.CkANNO.ckID);
  CkFORM.CkANNO.ckSize = (uint4_t) (4 + AFw.Hinfo.N);

  /* SSND chunk */
  MCOPY (CKID_SSND, CkFORM.CkSSND.ckID);
  CkFORM.CkSSND.ckSize = 8 + Ldata;
  CkFORM.CkSSND.offset = 0;
  CkFORM.CkSSND.blockSize = 0;

 /* Fill in the FORM chunk size */
  size = 4 + 8 + RNDUPV(CkFORM.CkCOMM.ckSize, ALIGN)
           + 8 + RNDUPV(CkFORM.CkFVER.ckSize, ALIGN)
           + 8 + RNDUPV(CkFORM.CkSSND.ckSize, ALIGN);
  if (AFw.Hinfo.N > 0)
    size += 8 + RNDUPV(CkFORM.CkANNO.ckSize, ALIGN);
  CkFORM.ckSize = (uint4_t) size;

/* Write out the header */
  WHEAD_S (fp, CkFORM.ckID);
  WHEAD_V (fp, CkFORM.ckSize, DS_EB);
  WHEAD_S (fp, CkFORM.AIFFID);

  WHEAD_S (fp, CkFORM.CkCOMM.ckID);
  WHEAD_V (fp, CkFORM.CkCOMM.ckSize, DS_EB);
  WHEAD_V (fp, CkFORM.CkCOMM.numChannels, DS_EB);
  WHEAD_V (fp, CkFORM.CkCOMM.numSampleFrames, DS_EB);
  WHEAD_V (fp, CkFORM.CkCOMM.sampleSize, DS_EB);
  WHEAD_S (fp, CkFORM.CkCOMM.sampleRate);
  WHEAD_S (fp, CkFORM.CkCOMM.compressionType);
  AF_wPstring (fp, CkFORM.CkCOMM.compressionName);
  WRPAD (fp, CkFORM.CkCOMM.ckSize, ALIGN);

  WHEAD_S (fp, CkFORM.CkFVER.ckID);
  WHEAD_V (fp, CkFORM.CkFVER.ckSize, DS_EB);
  WHEAD_V (fp, CkFORM.CkFVER.timestamp, DS_EB);

  if (AFw.Hinfo.N > 0) {
    WHEAD_S (fp, CkFORM.CkANNO.ckID);
    WHEAD_V (fp, CkFORM.CkANNO.ckSize, DS_EB);
    WHEAD_SN (fp, FM_AFSP, (sizeof FM_AFSP) - 1);	/* Omit null char */
    WHEAD_SN (fp, AFw.Hinfo.Info, AFw.Hinfo.N);
    WRPAD (fp, CkFORM.CkANNO.ckSize, ALIGN);
  }

  WHEAD_S (fp, CkFORM.CkSSND.ckID);
  WHEAD_V (fp, CkFORM.CkSSND.ckSize, DS_EB);
  WHEAD_V (fp, CkFORM.CkSSND.offset, DS_EB);
  WHEAD_V (fp, CkFORM.CkSSND.blockSize, DS_EB);

/* Set the parameters for file access */
  AFp = AFsetWrite (fp, FT_AIFF_C, &AFw);

  return AFp;
}

/* Set the scale factor */

static double
AI_ScaleF (int Format)

{
  double ScaleF;

  switch (Format) {
  case FD_INT8:
    ScaleF = 1./AI_SF_NONE8;
    break;
  case FD_INT16:
    ScaleF = 1./AI_SF_NONE16;
    break;
  case FD_INT24:
    ScaleF = 1./AI_SF_NONE24;
    break;
  case FD_INT32:
    ScaleF = 1./AI_SF_NONE32;
    break;
  case FD_MULAW8:
    ScaleF = 1./AI_SF_ULAW;
    break;
  case FD_ALAW8:
    ScaleF = 1./AI_SF_ALAW;
    break;
  case FD_FLOAT32:
    ScaleF = 1./AI_SF_FLOAT32;
    break;
  case FD_FLOAT64:
    ScaleF = 1./AI_SF_FLOAT64;
    break;
  default:
    UTwarn ("AFwrAIhead - %s", AFM_AIFF_UnsData);
    ScaleF = 0.0;
  }

  return ScaleF;
}

/* Fill in the COMM chunk */


static void
AF_setCOMM (struct AI_CkCOMM *CkCOMM, int Format, long int Nframe,
	    long int Nchan, double Sfreq)

{
  int ncP;

  MCOPY (CKID_COMM, CkCOMM->ckID);

/* Set up the encoding parameters */
  switch (Format) {
  case FD_INT8:
  case FD_INT16:
  case FD_INT24:
  case FD_INT32:
    MCOPY (CT_NONE, CkCOMM->compressionType);
    strcpy (CkCOMM->compressionName, CN_NONE);
    CkCOMM->sampleSize = 8 * AF_DL[Format];
    break; 

  /* There are two codes each for mu-law and A-law coding.  Until recently,
     the codes "ULAW" and "ALAW" originating with SGI software were used.
     The new Apple QuickTime software only recognizes "ulaw" and "alaw".
     Here we use the QuickTime conventions.
  */
  case FD_MULAW8:
    MCOPY (CT_ULAW, CkCOMM->compressionType);
    strcpy (CkCOMM->compressionName, CN_ULAW);
    CkCOMM->sampleSize = 16;		/* Uncompressed data size in bits */
    break;
  case FD_ALAW8:
    MCOPY (CT_ALAW, CkCOMM->compressionType);
    strcpy (CkCOMM->compressionName, CN_ALAW);
    CkCOMM->sampleSize = 16;		/* Uncompressed data size in bits */
    break;
  case FD_FLOAT32:
    MCOPY (CT_FLOAT32, CkCOMM->compressionType);
    strcpy (CkCOMM->compressionName, CN_FLOAT32);
    CkCOMM->sampleSize = 8 * AF_DL[Format];
    if (! UTcheckIEEE ())
      UTwarn ("AFwrAUhead - %s", AFM_AIFF_NoIEEE);
    break;
  case FD_FLOAT64:
    MCOPY (CT_FLOAT64, CkCOMM->compressionType);
    strcpy (CkCOMM->compressionName, CN_FLOAT64);
    CkCOMM->sampleSize = 8 * AF_DL[Format];
    if (! UTcheckIEEE ())
      UTwarn ("AFwrAUhead - %s", AFM_AIFF_NoIEEE);
    break;
  default:
    assert (0);
  }

  CkCOMM->numChannels = (uint2_t) Nchan;
  if (Nframe == AF_NFRAME_UNDEF)
    CkCOMM->numSampleFrames = (uint4_t) 0;
  else
    CkCOMM->numSampleFrames = (uint4_t) Nframe;
  /* CkCOMM.sampleSize filled in above */
  UTeIEEE80 (Sfreq, CkCOMM->sampleRate);
  /* CkCOMM.compressionType filled in above */
  /* CkCOMM.compressionName filled in above */

  ncP = AF_wPstring (NULL, CkCOMM->compressionName);
  CkCOMM->ckSize = 22 + ncP;

  return;
}

/* Write a P-string, returning the full length */
/* P-string:
   - number of characters in string (stored in a byte)
   - string
   - trailing null to make the overall size even
*/


static int
AF_wPstring (FILE *fp, char string[])

{
  char slen;
  int nc, ncP;

  nc = strlen (string);
  ncP = RNDUPV (nc + 1, 2);

  if (fp != NULL) {
    slen = (char) nc;
    WHEAD_V (fp, slen, DS_EB);
    WHEAD_SN (fp, string, nc);
    WHEAD_SN (fp, NULL, ncP - (nc + 1));
  }

  return ncP;
}
