/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  AFILE *AFwrWVhead (FILE *fp, int Format, long int Nframe, long int Nchan,
                     double Sfreq)

Purpose:
  Write header information to a RIFF WAVE file

Description:
  This routine opens and writes header information to a RIFF WAVE format audio
  file.

  RIFF WAVE file:
   Offset Length Type    Contents
      0     4    char   "RIFF" file identifier
      4     4    int    Chunk length
      8     4    char   "WAVE" file identifier
     12     4    char   "fmt " chunk identifier
     16     4    int    Chunk length
     20     2    int      Audio data type
     22     2    int      Number of interleaved channels
     24     4    int      Sample rate
     28     4    int      Average bytes/sec
     32     2    int      Block align
     34     2    int      Data word length (bits)
     36     2    int      Extra info size (0) (only for non-PCM data)
      C     4    int    "fact" chunk identifier (only for non-PCM data)
    C+4     4    int    Chunk length
    C+8     4    int      Number of samples
    ...   ...    ...    ...
      D     4    char   "data" chunk identifier
    D+4     4    int    Chunk length
    D+8    ...   ...      Audio data

Parameters:
  <-  AFILE AFwrWVhead
      Audio file pointer for the audio file.  In case of error, the audio file
      pointer is set to NULL.
   -> FILE *fp
      File pointer for the audio file
   -> int Format
      Audio file data format code: FD_MULAW8, FD_ALAW8, FD_UINT8, etc.
   -> long int Nframe
      Number of sample frames.  This value is used to set the file header
      information for non-random access files for which the header cannot be
      updated when the file is closed.  If this value is AF_NFRAME_UNDEF, the
      file must be random access.
   -> long int Nchan
      Number of channels
   -> double Sfreq
      Sampling frequency

Author / revision:
  P. Kabal  Copyright (C) 2001

  $Revision: 1.52 $  $Date: 2001/04/10 01:24:04 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFwrWVhead.c 1.52 2001/04/10 AFsp-v5r2 $";

#include <assert.h>
#include <setjmp.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/nucleus.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#define AF_DATA_LENGTHS
#include <libtsp/AFpar.h>
#include <libtsp/WVpar.h>

#define ICEILV(n,m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define RNDUPV(n,m)	((m) * ICEILV (n, m))		/* Round up */

#define MCOPY(src,dest)		memcpy ((void *) (dest), \
					(const void *) (src), sizeof (dest))
#define WRPAD(fp,size,align) \
     AFwriteHead (fp, NULL, 1, (int) (RNDUPV(size, align) - (size)), \
		  DS_NATIVE);

#define ALIGN		2	/* Chunks padded out to a multiple of ALIGN */

/* setjmp / longjmp environment */
extern jmp_buf AFW_JMPENV;

static void
AF_setFMT (struct WV_CKfmt *CKfmt, int Format, long int Nchan, double Sfreq);
static int
AF_wrFMT (FILE *fp, const struct WV_CKfmt *CKfmt);
static double
WV_ScaleF (int Format);


AFILE *
AFwrWVhead (FILE *fp, int Format, long int Nframe, long int Nchan,
	    double Sfreq)

{
  AFILE *AFp;
  int Lw;
  long int size, Ldata;
  struct WV_CKRIFF CKRIFF;
  struct AF_write AFw;

/* Set the long jump environment; on error return a NULL */
  if (setjmp (AFW_JMPENV))
    return NULL;	/* Return from a header write error */

/* Set up the encoding parameters */
  AFw.DFormat.ScaleF = WV_ScaleF (Format);
  if (AFw.DFormat.ScaleF <= 0.0)
    return NULL;
  AFw.DFormat.Format = Format;
  AFw.DFormat.Swapb = DS_EL;
  AFw.Sfreq = Sfreq;
  AFw.Nchan = Nchan;

  Lw = AF_DL[Format];
  if (FLseekable (fp))
    Ldata = 0L;
  else if (Nframe != AF_NFRAME_UNDEF)
    Ldata = Nframe * Nchan * Lw;
  else {
    UTwarn ("AFwrWVhead - %s", AFM_WV_WRAccess);
    return NULL;
  }

  /* RIFF chunk */
  MCOPY ("RIFF", CKRIFF.ckID);
  /* defer filling in the chunk size */ 
  MCOPY ("WAVE", CKRIFF.WAVEID);

  /* fmt chunk */
  AF_setFMT (&CKRIFF.CKfmt, Format, Nchan, Sfreq);

  /* fact chunk */
  if (CKRIFF.CKfmt.wFormatTag != WAVE_FORMAT_PCM) {
    MCOPY ("fact", CKRIFF.CKfact.ckID);
    CKRIFF.CKfact.ckSize = (uint4_t) sizeof (CKRIFF.CKfact.dwSampleLength);
    CKRIFF.CKfact.dwSampleLength = (uint4_t) 0;
  }

  /* data chunk */
  MCOPY ("data", CKRIFF.CKdata.ckID);
  CKRIFF.CKdata.ckSize = (uint4_t) Ldata;

  /* Fill in the RIFF chunk size */
  size = 4 + 8 + RNDUPV(CKRIFF.CKfmt.ckSize, ALIGN) +
             8 + RNDUPV(CKRIFF.CKdata.ckSize, ALIGN);
  if (CKRIFF.CKfmt.wFormatTag != WAVE_FORMAT_PCM)
    size += 8 + RNDUPV(CKRIFF.CKfact.ckSize, ALIGN);
  CKRIFF.ckSize = (uint4_t) size;

/* Write out the header */
  WHEAD_S (fp, CKRIFF.ckID);
  WHEAD_V (fp, CKRIFF.ckSize, DS_EL);
  WHEAD_S (fp, CKRIFF.WAVEID);

  AF_wrFMT (fp, &CKRIFF.CKfmt);

  if (CKRIFF.CKfmt.wFormatTag != WAVE_FORMAT_PCM) {
    WHEAD_S (fp, CKRIFF.CKfact.ckID);
    WHEAD_V (fp, CKRIFF.CKfact.ckSize, DS_EL);
    WHEAD_V (fp, CKRIFF.CKfact.dwSampleLength, DS_EL);
  }

  WHEAD_S (fp, CKRIFF.CKdata.ckID);
  WHEAD_V (fp, CKRIFF.CKdata.ckSize, DS_EL);

/* Header information records */
  AFw.Hinfo = *(AFgenHinfo (Sfreq));

/* Set the parameters for file access */
  AFp = AFsetWrite (fp, FT_WAVE, &AFw);

  return AFp;
}
/* Set the scale factor */


static double
WV_ScaleF (int Format)

{
  double ScaleF;

  switch (Format) {
  case FD_INT16:
    ScaleF = 1./WV_SF_PCM16;
    break;
  case FD_INT24:
    ScaleF = 1./WV_SF_PCM24;
    break;
  case FD_INT32:
    ScaleF = 1./WV_SF_PCM32;
    break;
  case FD_UINT8:
    ScaleF = 1./WV_SF_PCM8;
    break;
  case FD_MULAW8:
    ScaleF = 1./WV_SF_MULAW;
    break;
  case FD_ALAW8:
    ScaleF = 1./WV_SF_ALAW;
    break;
  case FD_FLOAT32:
    ScaleF = 1./WV_SF_FLOAT32;
    if (! UTcheckIEEE ())
      UTwarn ("AFwrWVhead - %s", AFM_WV_NoIEEE);
    break;
  case FD_FLOAT64:
    ScaleF = 1./WV_SF_FLOAT64;
    if (! UTcheckIEEE ())
      UTwarn ("AFwrWVhead - %s", AFM_WV_NoIEEE);
    break;
  default:
    UTwarn ("AFwrWVhead - %s", AFM_WV_UnsData);
    ScaleF = 0.0;
  }

  return ScaleF;
}

/* Fill in the fmt chunk */


static void
AF_setFMT (struct WV_CKfmt *CKfmt, int Format, long int Nchan, double Sfreq)

{
  int Lw;

  MCOPY ("fmt ", CKfmt->ckID);
  CKfmt->ckSize = sizeof (CKfmt->wFormatTag) + sizeof (CKfmt->nChannels)
    + sizeof (CKfmt->nSamplesPerSec) + sizeof (CKfmt->nAvgBytesPerSec)
    + sizeof (CKfmt->nBlockAlign) + sizeof (CKfmt->wBitsPerSample);

  switch (Format) {
  case FD_UINT8:
  case FD_INT16:
  case FD_INT24:
  case FD_INT32:
    CKfmt->wFormatTag = WAVE_FORMAT_PCM;
    break;
  case FD_MULAW8:
    CKfmt->wFormatTag = WAVE_FORMAT_MULAW;
    break;
  case FD_ALAW8:
    CKfmt->wFormatTag = WAVE_FORMAT_ALAW;
    break;
  case FD_FLOAT32:
  case FD_FLOAT64:
    CKfmt->wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
    break;
  default:
    assert (0);
  }

  Lw = AF_DL[Format];
  CKfmt->nChannels = (uint2_t) Nchan;
  CKfmt->nSamplesPerSec = (uint4_t) (Sfreq + 0.5);	/* Rounding */
  CKfmt->nAvgBytesPerSec = (uint4_t) (CKfmt->nSamplesPerSec * Nchan * Lw);
  CKfmt->nBlockAlign = (uint2_t) (Lw * Nchan);
  CKfmt->wBitsPerSample = (uint2_t) (8 * Lw);
  if (CKfmt->wFormatTag != WAVE_FORMAT_PCM) {
    CKfmt->cbSize = (uint2_t) 0;
    CKfmt->ckSize += sizeof (CKfmt->cbSize);
  }

  return;
}

/* Write the format chunk */


static int
AF_wrFMT (FILE *fp, const struct WV_CKfmt *CKfmt)

{
  long int offs, LFMT;

  offs  = WHEAD_S (fp, CKfmt->ckID);
  offs += WHEAD_V (fp, CKfmt->ckSize, DS_EL);
  LFMT = CKfmt->ckSize + 8;
  offs += WHEAD_V (fp, CKfmt->wFormatTag, DS_EL);
  offs += WHEAD_V (fp, CKfmt->nChannels, DS_EL);
  offs += WHEAD_V (fp, CKfmt->nSamplesPerSec, DS_EL);
  offs += WHEAD_V (fp, CKfmt->nAvgBytesPerSec, DS_EL);
  offs += WHEAD_V (fp, CKfmt->nBlockAlign, DS_EL);
  offs += WHEAD_V (fp, CKfmt->wBitsPerSample, DS_EL);
  if (offs < LFMT)
    offs += WHEAD_V (fp, CKfmt->cbSize, DS_EL);

  assert (LFMT == offs);

  return offs;
}
